#!/usr/bin/python3
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.metrics.pairwise import cosine_similarity
from responses import responses_dict as responses
from nltk.tokenize import word_tokenize
from nltk.stem import WordNetLemmatizer
from nltk.corpus import stopwords
import numpy as np
import speech_recognition as sr
import pyttsx3
import requests
import os
import nltk
nltk.download('stopwords')
# nltk.download('punkt')
# nltk.download('wordnet')


class VoiceAssistant:
    def __init__(self):
        self.recognizer = sr.Recognizer()
        self.microphone = sr.Microphone()
        self.speaker = pyttsx3.init()
        self.speaker.setProperty('voice', 'HKEY_LOCAL_MACHINE\SOFTWARE\Microsoft\Speech\Voices\Tokens\TTS_MS_EN-US_DAVID_11.0')
        self.speaker.setProperty('rate', 150)
        self.speaker.setProperty('volume', 1.0)
        self.lemmatizer = WordNetLemmatizer()
        self.stop_words = set(stopwords.words('english'))

    def speak(self, text):
        self.speaker.say(text)
        self.speaker.runAndWait()

    def listen(self):
        with self.microphone as source:
            print("Say something!")
            self.recognizer.adjust_for_ambient_noise(source)
            audio = self.recognizer.listen(source)
            try:
                text = self.recognizer.recognize_google(audio, language='en-US')
                print("Google Assistant: thinks you said: " + text)
                if text == None:
                    text = self.recognizer.recognize_sphinx(audio, language='en-US')
                    print("Sphinx Assistant: thinks you said: " + text)
                assistant.speak("I heard: " + text)
                return text
            except sr.UnknownValueError:
                print("Assistant could not understand audio")
            except sr.RequestError as e:
                print("Assistant error; {0}".format(e))

    def preprocess_text(self, text):
        if text:
            tokens = word_tokenize(text)
            words = [self.lemmatizer.lemmatize(word) for word in tokens if word.isalpha() and word not in self.stop_words]
            return ' '.join(words)
        return ""
    
    def get_response(self, query, source=responses):
        if query:
            try:
                # Preprocess the query
                preprocessed_query = self.preprocess_text(query)
                tfidf_vectorizer = TfidfVectorizer()
                answers = list(source.keys())
                tfidf_matrix = tfidf_vectorizer.fit_transform(answers)
                query_vec = tfidf_vectorizer.transform([preprocessed_query])
                print(f"Query vector: {query_vec}")
                # Calculate cosine similarity between query and answers
                cosine_similarities = np.dot(query_vec, tfidf_matrix.T)
                most_similar_idx = np.argmax(cosine_similarities)
                print(f"Cosine similarity: {cosine_similarities}\nMost similar idx: {most_similar_idx}")
                response_key = answers[most_similar_idx]
                response = source[response_key]
                while callable(response):
                    callable_response = response(self)

                    if callable_response == False:
                        self.speak("Sorry, I couldn't process this information.")
                    else:
                        return callable_response
                return response
            except AttributeError:
                return "I'm sorry, I didn't understand that."
            
    def search_article(self, query, sentences):
        vectorizer = TfidfVectorizer()
        preprocessed_sentences = [self.preprocess_text(sentence) for sentence in sentences]
        tfidf_matrix = vectorizer.fit_transform(preprocessed_sentences)
        preprocessed_query = self.preprocess_text(query)
        query_tfidf = vectorizer.transform([preprocessed_query])
        similarities = cosine_similarity(query_tfidf, tfidf_matrix)
        best_match_idx = np.argmax(similarities, axis=1)[0]
        return sentences[best_match_idx]

def help_listen(voice):
    voice_path =  os.path.join(os.getcwd(), voice)
    with sr.AudioFile(voice_path) as source:
        audio = assistant.recognizer.record(source)
        text = assistant.recognizer.recognize_google(audio, language='en-US')
        print("Google Assistant: thinks you said: " + text)
        if text == None:
            text = assistant.recognizer.recognize_sphinx(audio, language='en-US')
            print("Sphinx Assistant: thinks you said: " + text)
        assistant.speak("I heard: " + text)
        response = assistant.get_response(text)
        assistant.speak(response)


if __name__ == "__main__":
    try:
        assistant = VoiceAssistant()
        #Test voice assistant
        #"wikipedia.wav", 
        voice_list = ["hi.wav", "what_time.wav", "your_name.wav", "random_cat.wav", "by.wav"]
        for voice in voice_list:
            print(voice)
            help_listen(voice)


        # Live voice assistant
        # while True:
        #     text = assistant.listen()
        #     if text:
        #         response = assistant.get_response(text)
        #         assistant.speak(response)
    except KeyboardInterrupt:
        print("Assistant stopped")
        exit(0)